#include <stdio.h>
#include <stdlib.h>
#include <pthread.h>

#if defined(__linux__)
#include <signal.h>
#include <sys/timeb.h>
#else
#include <windows.h>
#include <time.h>
#endif

#include "ini.h"
#include "rs232.h"

typedef struct
{
    char *port_path;
    unsigned int send_byte_once;
    unsigned int send_loop_cnt;
    unsigned int sleep_interval;
    unsigned int baud_rate;

} INI_ITEM;


INI_ITEM ini_item;

#if defined(__linux__)
    int port_fd = -1;
#else
    HANDLE port_hd = NULL;
#endif

char *start_time = NULL;
char *end_time = NULL;

void initDev (void);
void startSend (void);
void getIni (void);
int handlerIni (void* user, const char* section, const char* name, const char* value);

char *get_now_time (void);
char *cal_itvl (char *start_time, char *end_time);
void split (char **arr, char *str, const char *del);

/*------------------------------------------------------------------*/
int main (void)
{
    getIni();
    initDev();
    startSend();

    return 0;
}


/*------------------------------------------------------------------*/
void initDev (void)
{
    char *port_path = ini_item.port_path;
    int  baud_rate = ini_item.baud_rate;
    char mode[] = {'8', 'N', '1', 0};


    #if defined(__linux__)
    port_fd = RS232_OpenComport(port_path, baud_rate, mode);
    if(port_fd == -1)
    {
        printf("Open '%s' fail\n", port_path);
        exit(1);
    }
    RS232_flushRX(port_fd);
    #else
    port_hd = RS232_OpenComport(port_path, baud_rate, mode);
    if(port_hd == NULL)
    {
        printf("Open '%s' fail\n", port_path);
        exit(1);
    }
    RS232_flushRX(port_hd);
    #endif

    printf("Open '%s' success\n", port_path);
}


/*------------------------------------------------------------------*/
void startSend (void)
{
    unsigned int send_byte_once = ini_item.send_byte_once;
    unsigned int send_loop_cnt = ini_item.send_loop_cnt;
    unsigned int sleep_interval = ini_item.sleep_interval;
    unsigned char *buf_send = (unsigned char *) malloc(send_byte_once * sizeof(unsigned char));

    for(unsigned int i=0; i<send_byte_once; i++)
    {
        buf_send[i] = i;
    }

    start_time = get_now_time();

    for(int i=0; i<send_loop_cnt; i++)
    {
        #if defined(__linux__)
        if(RS232_SendBuf(port_fd, buf_send, send_byte_once) == send_byte_once)
        #else
        if(RS232_SendBuf(port_hd, buf_send, send_byte_once) == send_byte_once)
        #endif
        {
            printf("Send success ... (%2d/%d)\n", i+1, send_loop_cnt);
        }
        else
        {
            printf("Send fail ... (%2d/%d)\n", i+1, send_loop_cnt);
        }

        if(i < send_loop_cnt - 1)
        {
            #if defined(__linux__)
            usleep(sleep_interval);
            #else
            Sleep(sleep_interval);
            #endif
        }
    }

    end_time = get_now_time();

    printf("\n");
    printf("Start time: %s\n", start_time);
    printf("End time:   %s\n", end_time);
    printf("Interval:   %s\n", cal_itvl(start_time, end_time));  
}


/*------------------------------------------------------------------*/
void getIni (void)
{
    if(ini_parse("setup.ini", handlerIni, &ini_item) < 0)
    {
        printf("Open setup.ini failed\n");
        exit(1);
    }
}


/*------------------------------------------------------------------*/
int handlerIni (void* user, const char* section, const char* name, const char* value)
{
    #define MATCH(s, n) strcmp(section, s) == 0 && strcmp(name, n) == 0

    INI_ITEM *p_ini = (INI_ITEM *) user;

    /* [setting] section */
    if(MATCH("setting", "port_path"))
    {
        p_ini->port_path = strdup(value);
    }
    else if(MATCH("setting", "send_byte_once"))
    {
        p_ini->send_byte_once = (unsigned int) atoi(value);
    }
    else if(MATCH("setting", "send_loop_cnt"))
    {
        p_ini->send_loop_cnt = (unsigned int) atoi(value);
    }
    else if(MATCH("setting", "baud_rate"))
    {
        p_ini->baud_rate = (unsigned int) atoi(value);
    }
    else if(MATCH("setting", "sleep_interval"))
    {
        #if defined(__linux__)
        p_ini->sleep_interval = (unsigned int) atoi(value);
        #else
        p_ini->sleep_interval = ((unsigned int) atoi(value)) / 1000;
        #endif
    }
    /* unknown section or name*/
    else
    {
        return 0;
    }

    return 1;

}


/*------------------------------------------------------------------*/
char *get_now_time (void)
{
    #define  TIME_CHAR_NUM  13

    char          time_str[TIME_CHAR_NUM];
    struct tm    *timeinfo;
    struct timeb  now;

    ftime(&now);
    timeinfo = localtime((time_t*) &now);
    sprintf(time_str, "%02d:%02d:%02d:%03d", timeinfo->tm_hour, timeinfo->tm_min, timeinfo->tm_sec, now.millitm);

    return strdup(time_str);
}


/*------------------------------------------------------------------*/
char *cal_itvl (char *start_time, char *end_time)
{
    int i;
    int j;
    const char *del = ":";
    char rtn_str[12 + 1];

    /* 4: hour/min/sec/ms */
    char *start_arr[4];
    char *end_arr[4];
    int start_int[4];
    int end_int[4];
    int itvl_int[4] = {0};

    split(start_arr, start_time, del);
    split(end_arr, end_time, del);

    i = 0;
    j = 0;
    while(i<4 && j<4)
    {
        start_int[i++] = atoi(*(start_arr+j++));
    }

    i = 0;
    j = 0;
    while(i<4 && j<4)
    {
        end_int[i++] = atoi(*(end_arr+j++));
    }

    /* now we get hour/min/sec/ms of both start time and end time */
    if(end_int[0] >= start_int[0])
    {
        /* ms */
        if(end_int[3] >= start_int[3])
        {
            itvl_int[3] = end_int[3] - start_int[3];
        }
        else
        {
            itvl_int[3] = end_int[3] + 1000 - start_int[3];
            end_int[2] = end_int[2] - 1;
        }

        /* sec */
        if(end_int[2] >= start_int[2])
        {
            itvl_int[2] = end_int[2] - start_int[2];
        }
        else
        {
            itvl_int[2] = end_int[2] + 60 - start_int[2];
            end_int[1] = end_int[1] - 1;
        }

        /* min */
        if(end_int[1] >= start_int[1])
        {
            itvl_int[1] = end_int[1] - start_int[1];
        }
        else
        {
            itvl_int[1] = end_int[1] + 60 - start_int[1];
            end_int[0] = end_int[0] - 1;
        }
    }
    /* the next day */
    else
    {
        int total_start_int[4] = {0};
        int total_end_int[4] = {0};

        /* get start total itvl */
        if(start_int[3] != 0)
        {
            total_start_int[3] = 1000 - start_int[3];
            start_int[2]++;
        }
        else
        {
            total_start_int[3] = 0;
        }

        if(start_int[2] != 0)
        {
            total_start_int[2] = 60 - start_int[2];
            start_int[1]++;
        }
        else
        {
            total_start_int[2] = 0;
        }

        if(start_int[1] != 0)
        {
            total_start_int[1] = 60 - start_int[1];
            start_int[0]++;
        }
        else
        {
            total_start_int[1] = 0;
        }

        total_start_int[0] = 24 - start_int[0];


        /* get end total itvl */
        for(int i=0; i<4; i++)
        {
            total_end_int[i] = end_int[i];
        }

        /* get all itvl */
        for(int i=0; i<4; i++)
        {
            itvl_int[i] = total_start_int[i] + total_end_int[i];
        }

        if(itvl_int[3] >= 1000)
        {
            itvl_int[3] = itvl_int[3] - 1000;
            itvl_int[2]++;
        }

        if(itvl_int[2] >= 60)
        {
            itvl_int[2] = itvl_int[2] - 60;
            itvl_int[1]++;
        }

        if(itvl_int[1] >= 60)
        {
            itvl_int[1] = itvl_int[1] - 60;
            itvl_int[0]++;
        }
    }

    sprintf(rtn_str, "%02d:%02d:%02d:%03d", itvl_int[0], itvl_int[1], itvl_int[2], itvl_int[3]);
    return strdup(rtn_str);
}


/*----------------------------------------------------------------------*/
void split (char **arr, char *str, const char *del)
{
    char *s = strtok(str, del);

    while(s != NULL)
    {
        *arr++ = s;
        s = strtok(NULL, del);
    }
}
